# Copyright 2021 c00k1ez (https://github.com/c00k1ez). All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable, Dict, Optional, Tuple, Union

import torch
import torch.nn as nn
import torch.nn.functional as F
from plain_transformers.losses import LabelSmoothingLoss


# TODO: write more complex solution for embedding sharing
class MultimodalTransformer(nn.Module):
    def __init__(
        self,
        first_encoder_class: Callable,
        second_encoder_class: Callable,
        decoder_class: Callable,
        d_model: int,
        first_encoder_vocab_size: int,
        first_encoder_max_length: int,
        first_encoder_pad_token_id: int,
        first_encoder_token_type_vocab_size: int,
        first_encoder_n_heads: int,
        first_encoder_dim_feedforward: int,
        first_encoder_num_layers: int,
        second_encoder_vocab_size: int,
        second_encoder_max_length: int,
        second_encoder_pad_token_id: int,
        second_encoder_token_type_vocab_size: int,
        second_encoder_n_heads: int,
        second_encoder_dim_feedforward: int,
        second_encoder_num_layers: int,
        decoder_max_length: int,
        decoder_vocab_size: int,
        decoder_pad_token_id: int,
        decoder_token_type_vocab_size: int,
        decoder_n_heads: int,
        decoder_dim_feedforward: int,
        decoder_num_layers: int,
        decoder_use_embedding_layer_norm: Optional[bool] = True,
        decoder_pos_embedding_type: Optional[str] = "embedding",
        decoder_activation_name: Optional[str] = "gelu",
        decoder_ln_eps: Optional[float] = 1e-12,
        decoder_dropout: Optional[float] = 0.1,
        decoder_layerdrop_threshold: Optional[float] = 0.0,
        decoder_type: Optional[str] = "post_ln",
        first_encoder_dropout: Optional[float] = 0.1,
        first_encoder_use_embedding_layer_norm: Optional[bool] = True,
        first_encoder_pos_embedding_type: Optional[str] = "embedding",
        first_encoder_activation_name: Optional[str] = "gelu",
        first_encoder_ln_eps: Optional[float] = 1e-12,
        first_encoder_layerdrop_threshold: Optional[float] = 0.0,
        first_encoder_type: Optional[str] = "post_ln",
        second_encoder_dropout: Optional[float] = 0.1,
        second_encoder_use_embedding_layer_norm: Optional[bool] = True,
        second_encoder_pos_embedding_type: Optional[str] = "embedding",
        second_encoder_activation_name: Optional[str] = "gelu",
        second_encoder_ln_eps: Optional[float] = 1e-12,
        second_encoder_layerdrop_threshold: Optional[float] = 0.0,
        second_encoder_type: Optional[str] = "post_ln",
        share_decoder_head_weights: Optional[bool] = True,
        share_encoder_decoder_embeddings: Optional[bool] = False,
        share_encoder_embeddings: Optional[bool] = False,
        label_smoothing: Optional[float] = 0.0,
    ) -> None:
        super(MultimodalTransformer, self).__init__()
        self.first_encoder = first_encoder_class(
            d_model=d_model,
            max_length=first_encoder_max_length,
            pad_token_id=first_encoder_pad_token_id,
            token_type_vocab_size=first_encoder_token_type_vocab_size,
            n_heads=first_encoder_n_heads,
            dim_feedforward=first_encoder_dim_feedforward,
            num_layers=first_encoder_num_layers,
            dropout=first_encoder_dropout,
            use_embedding_layer_norm=first_encoder_use_embedding_layer_norm,
            pos_embedding_type=first_encoder_pos_embedding_type,
            activation_name=first_encoder_activation_name,
            ln_eps=first_encoder_ln_eps,
            vocab_size=first_encoder_vocab_size,
            layerdrop_threshold=first_encoder_layerdrop_threshold,
            encoder_type=first_encoder_type,
        )
        self.second_encoder = second_encoder_class(
            d_model=d_model,
            max_length=second_encoder_max_length,
            pad_token_id=second_encoder_pad_token_id,
            token_type_vocab_size=second_encoder_token_type_vocab_size,
            n_heads=second_encoder_n_heads,
            dim_feedforward=second_encoder_dim_feedforward,
            num_layers=second_encoder_num_layers,
            dropout=second_encoder_dropout,
            use_embedding_layer_norm=second_encoder_use_embedding_layer_norm,
            pos_embedding_type=second_encoder_pos_embedding_type,
            activation_name=second_encoder_activation_name,
            ln_eps=second_encoder_ln_eps,
            vocab_size=second_encoder_vocab_size,
            layerdrop_threshold=second_encoder_layerdrop_threshold,
            encoder_type=second_encoder_type,
        )

        self.decoder = decoder_class(
            d_model=d_model,
            vocab_size=decoder_vocab_size,
            max_length=decoder_max_length,
            pad_token_id=decoder_pad_token_id,
            token_type_vocab_size=decoder_token_type_vocab_size,
            n_heads=decoder_n_heads,
            dim_feedforward=decoder_dim_feedforward,
            num_layers=decoder_num_layers,
            dropout=decoder_dropout,
            use_embedding_layer_norm=decoder_use_embedding_layer_norm,
            pos_embedding_type=decoder_pos_embedding_type,
            activation_name=decoder_activation_name,
            ln_eps=decoder_ln_eps,
            layerdrop_threshold=decoder_layerdrop_threshold,
            decoder_type=decoder_type,
        )
        self.lm_head = nn.Linear(d_model, decoder_vocab_size, bias=False)
        if share_decoder_head_weights:
            self.lm_head.weight = self.decoder.embedding.token_embedding.weight
        if share_encoder_decoder_embeddings:
            self.first_encoder.embedding.token_embedding.weight = self.decoder.embedding.token_embedding.weight
            self.second_encoder.embedding.token_embedding.weight = self.decoder.embedding.token_embedding.weight
        if share_encoder_embeddings:
            self.first_encoder.embedding.token_embedding.weight = self.second_encoder.embedding.token_embedding.weight
        self.loss_function = LabelSmoothingLoss(
            smoothing=label_smoothing,
            ignore_index=decoder_pad_token_id,
        )
        self.pad_token_id = decoder_pad_token_id

    def forward(
        self,
        first_encoder_input_ids: torch.Tensor,
        second_encoder_input_ids: torch.Tensor,
        labels: torch.Tensor,
        decoder_attention_mask: Optional[torch.Tensor] = None,
        first_encoder_attention_mask: Optional[torch.Tensor] = None,
        second_encoder_attention_mask: Optional[torch.Tensor] = None,
        get_attention_scores: Optional[bool] = False,
        cached_encoder_state: Optional[Tuple[Dict[str, torch.Tensor], Dict[str, torch.Tensor]]] = None,
        return_encoder_state: Optional[bool] = False,
        compute_loss: Optional[bool] = False,
        get_logits: Optional[bool] = False,
    ) -> Dict[str, Union[torch.Tensor, Dict[str, torch.Tensor], Dict[str, Dict[str, torch.Tensor]],],]:
        first_encoder_state, second_encoder_state = None, None
        attn_scores = {}
        if cached_encoder_state is not None:
            first_encoder_state, second_encoder_state = cached_encoder_state
        else:
            first_encoder_state = self.first_encoder(
                input_ids=first_encoder_input_ids,
                attention_mask=first_encoder_attention_mask,
                token_type_ids=None,
                get_attention_scores=get_attention_scores,
            )

            second_encoder_state = self.second_encoder(
                input_ids=second_encoder_input_ids,
                attention_mask=second_encoder_attention_mask,
                token_type_ids=None,
                get_attention_scores=get_attention_scores,
            )
            if get_attention_scores:
                attn_scores["encoder"] = {
                    "first_encoder": first_encoder_state[1],
                    "second_encoder": second_encoder_state[1],
                }

            first_encoder_state = {
                "key": first_encoder_state[0],
                "value": first_encoder_state[0],
            }
            second_encoder_state = {
                "key": second_encoder_state[0],
                "value": second_encoder_state[0],
            }
        hidden = self.decoder(
            input_ids=labels,
            first_encoder_hidden_state=first_encoder_state,
            second_encoder_hidden_state=second_encoder_state,
            attention_mask=decoder_attention_mask,
            first_encoder_attention_mask=first_encoder_attention_mask,
            second_encoder_attention_mask=second_encoder_attention_mask,
            get_attention_scores=get_attention_scores,
        )
        if get_attention_scores:
            attn_scores["decoder"] = hidden[1]
        hidden = hidden[0]
        raw_probs = self.lm_head(hidden)

        output = {"lm_probs": torch.softmax(raw_probs, dim=-1) if not get_logits else raw_probs}
        if return_encoder_state:
            output["encoder_hidden_state"] = (
                first_encoder_state,
                second_encoder_state,
            )
        if compute_loss:
            batch_size = labels.shape[0]
            labels = torch.cat(
                [
                    labels[:, 1:],
                    torch.LongTensor([[self.pad_token_id]]).to(labels.device).repeat(batch_size, 1),
                ],
                dim=-1,
            )
            output["loss_val"] = self.loss_function(raw_probs.view(-1, raw_probs.shape[-1]), labels.view(-1))
        return output



# Copyright 2021 c00k1ez (https://github.com/c00k1ez). All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch

from .base_sampler import BaseSampler


class NucleusSampler(BaseSampler):
    """
    Sampling based on
    https://gist.github.com/thomwolf/1a5a29f6962089e871b94cbd09daf317
    """

    def __init__(self, *args, **kwargs) -> None:
        super(NucleusSampler, self).__init__(*args, **kwargs)

    @torch.no_grad()
    def sample(self, logits: torch.Tensor, temperatype: float, **kwargs) -> torch.Tensor:
        # logits shape: [1, vocab_size]
        top_k, top_p = logits.shape[-1], 0.0
        if "top_k" in kwargs:
            top_k = kwargs["top_k"]
        if "top_p" in kwargs:
            top_p = kwargs["top_p"]
        logits = logits / temperatype
        if top_k > 0:
            inds_to_delete = logits < torch.topk(logits, top_k)[0][..., -1, None]
            logits[inds_to_delete] = float("-inf")

        if top_p > 0.0:
            sorted_logits, sorted_indices = torch.sort(logits, descending=True)
            cumulative_probs = torch.cumsum(torch.softmax(sorted_logits, dim=-1), dim=-1)
            sorted_indices_to_remove = cumulative_probs >= top_p
            sorted_indices_to_remove[..., 1:] = sorted_indices_to_remove[..., :-1].clone()
            sorted_indices_to_remove[..., 0] = 0
            indices_to_remove = torch.zeros_like(logits, dtype=sorted_indices_to_remove.dtype).scatter_(
                dim=-1, index=sorted_indices, src=sorted_indices_to_remove
            )  # .to(logits.device)
            logits[indices_to_remove] = float("-inf")

        return torch.multinomial(torch.softmax(logits, dim=-1), 1)



# Copyright 2021 c00k1ez (https://github.com/c00k1ez). All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
import random
from typing import Callable, Optional, Tuple, Union

import torch
import torch.nn as nn
import torch.nn.functional as F

from .utils import act_to_func


class FFN(nn.Module):
    def __init__(
        self,
        d_model: int,
        dim_feedforward: int,
        dropout: Optional[float] = 0.1,
        activation_name: Optional[str] = "gelu",
    ) -> None:
        super(FFN, self).__init__()
        self.d_model = d_model
        self.activation_name = activation_name
        self.dim_feedforward = dim_feedforward
        self.layer_inc = nn.Linear(d_model, dim_feedforward)
        self.layer_reduce = nn.Linear(dim_feedforward, d_model)
        self.dropout = nn.Dropout(dropout)

    def forward(self, hidden: torch.Tensor) -> torch.Tensor:
        hid_state = self.layer_inc(hidden)
        hid_state = self.dropout(act_to_func(self.activation_name)(hid_state))
        return self.layer_reduce(hid_state)


class MultiHeadAttention(nn.Module):
    def __init__(
        self,
        d_model: int,
        n_heads: int,
        dropout: Optional[float] = 0.1,
        query_input_dim: Optional[int] = None,
        key_input_dim: Optional[int] = None,
        value_input_dim: Optional[int] = None,
        context_len: Optional[int] = None,
    ) -> None:
        super(MultiHeadAttention, self).__init__()
        assert d_model % n_heads == 0
        self.d_model = d_model
        self.n_heads = n_heads
        self.hidden_per_head = d_model // n_heads
        self.scale = self.hidden_per_head ** 0.5
        self.query_input_dim = d_model if query_input_dim is None else query_input_dim
        self.key_input_dim = d_model if key_input_dim is None else key_input_dim
        self.value_input_dim = d_model if value_input_dim is None else value_input_dim

        self.key_projection = nn.Linear(self.key_input_dim, d_model)
        self.query_projection = nn.Linear(self.query_input_dim, d_model)
        self.value_projection = nn.Linear(self.value_input_dim, d_model)
        self.dropout = nn.Dropout(dropout)

        # attn_type in ["encoder", "decoder"]
        self.attn_type = "encoder"

        if context_len is not None:
            self.attn_type = "decoder"
            self.register_buffer("masked_val", torch.FloatTensor([-1e4]))
            self.register_buffer(
                "tri_mask",
                torch.tril(torch.ones((context_len, context_len), dtype=torch.uint8)).view(
                    1, 1, context_len, context_len
                ),
            )
        else:
            self.register_buffer("masked_val", None)
            self.register_buffer("tri_mask", None)

    def _transpose_to_heads(self, x: torch.Tensor) -> torch.Tensor:
        # (batch_size, seq_len, emb_dim) ->
        # (batch_size, seq_len, n_heads, hidden_per_head)
        new_shape = x.shape[:-1] + (self.n_heads, self.hidden_per_head)
        x = x.view(*new_shape)
        # (batch_size, seq_len, n_heads, hidden_per_head) ->
        # -> (batch_size, n_heads, seq_len, hidden_per_head)
        return x.permute(0, 2, 1, 3)

    def _generate_decoder_self_attn_mask(self, q_seq_len: int, k_seq_len: int) -> torch.Tensor:
        # TODO: fix case then k_seq_len < q_seq_len
        if self.training:
            attn_mask = self.tri_mask[:, :, k_seq_len - q_seq_len : k_seq_len, :k_seq_len]
        else:
            attn_mask = torch.ones((1, 1, k_seq_len, k_seq_len)).type_as(self.tri_mask)
        return attn_mask

    def forward(
        self,
        query: torch.Tensor,
        key: torch.Tensor,
        value: torch.Tensor,
        attention_mask: Optional[torch.Tensor] = None,
        get_attention_scores: Optional[bool] = False,
    ) -> Union[Tuple[torch.Tensor], Tuple[torch.Tensor, torch.Tensor]]:
        query_proj = self.query_projection(query)
        key_proj = self.key_projection(key)
        value_proj = self.value_projection(value)

        query_proj = self._transpose_to_heads(query_proj)
        key_proj = self._transpose_to_heads(key_proj)
        value_proj = self._transpose_to_heads(value_proj)

        raw_scores = torch.matmul(query_proj, key_proj.transpose(-1, -2))

        if self.attn_type == "decoder":
            decoder_attn_mask = self._generate_decoder_self_attn_mask(key_proj.shape[2], key_proj.shape[2])
            raw_scores = torch.where(
                decoder_attn_mask.bool(),
                raw_scores,
                self.masked_val.to(raw_scores.dtype),
            )

        if attention_mask is not None:
            raw_scores = raw_scores + attention_mask
        attn_scores = F.softmax(raw_scores / self.scale, dim=-1)
        attn = torch.matmul(attn_scores, value_proj)
        attn = attn.permute(0, 2, 1, 3).contiguous()
        new_shape = attn.shape[:-2] + (self.d_model,)
        attn = attn.view(*new_shape)
        output = (attn,)
        if get_attention_scores:
            output = output + (attn_scores,)
        return output


class SinusoidalPositionalEmbedding(nn.Module):
    # based on implenemtation from fairseq
    def __init__(self, context_len: int, embedding_dim: int):
        super(SinusoidalPositionalEmbedding, self).__init__()
        # im not shure about +1 operation
        self.context_len = context_len + 1
        self.embedding_dim = embedding_dim

        half_dim = embedding_dim // 2
        weight = math.log(10000) / (half_dim - 1)
        weight = torch.exp(torch.arange(half_dim, dtype=torch.float) * -weight)
        weight = torch.arange(self.context_len, dtype=torch.float).unsqueeze(1) * weight.unsqueeze(0)
        weight = torch.cat([torch.sin(weight), torch.cos(weight)], dim=1).view(self.context_len, -1)
        if embedding_dim % 2 == 1:
            weight = torch.cat([weight, torch.zeros(self.context_len, 1)], dim=1)
        self.weight = nn.Parameter(weight, requires_grad=False)

    def forward(
        self,
        input_positions: torch.Tensor,
    ) -> torch.Tensor:
        batch_size, seq_len = input_positions.shape
        # im not shure about +1 operation
        # TODO: find more information about it
        input_positions = input_positions + 1
        inds = self.weight.index_select(0, input_positions.view(-1)).view(batch_size, seq_len, -1).detach()
        return inds


class TransformerEmbedding(nn.Module):
    def __init__(
        self,
        vocab_size: int,
        d_model: int,
        max_length: int,
        pad_token_id: int,
        token_type_vocab_size: int,
        pos_embedding_type: Optional[str] = "embedding",
        dropout: Optional[float] = 0.1,
        use_layer_norm: Optional[bool] = False,
        ln_eps: Optional[float] = 1e-12,
    ) -> None:
        super(TransformerEmbedding, self).__init__()
        assert token_type_vocab_size >= 0
        assert vocab_size >= 0
        assert d_model >= 0
        assert max_length > 0
        if token_type_vocab_size == 0:
            use_token_type_embeddings = False
        else:
            use_token_type_embeddings = True
        assert pos_embedding_type in ["embedding", "timing"]

        self.token_embedding = nn.Embedding(vocab_size, d_model, pad_token_id)

        if pos_embedding_type == "embedding":
            self.positional_embedding = nn.Embedding(max_length, d_model)
        else:
            self.positional_embedding = SinusoidalPositionalEmbedding(max_length, d_model)
        if use_token_type_embeddings:
            self.token_type_embedding = nn.Embedding(token_type_vocab_size, d_model)
        else:
            self.token_type_embedding = nn.Identity()
            self.token_type_embedding.register_parameter("weight", None)

        self.dropout = nn.Dropout(dropout)
        self.register_buffer("pos_ids", torch.arange(max_length).expand((1, -1)))

        if use_layer_norm:
            self.layer_norm = nn.LayerNorm(d_model, eps=ln_eps)
        else:
            self.layer_norm = nn.Identity()
            self.layer_norm.register_parameter("weight", None)
            self.layer_norm.register_parameter("bias", None)

        self.use_layer_norm = use_layer_norm
        self.use_token_type_embeddings = use_token_type_embeddings

    def forward(
        self,
        input_ids: torch.Tensor,
        token_type_ids: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        token_emb = self.token_embedding(input_ids)
        input_shape = token_emb.shape
        if token_type_ids is None:
            token_type_ids = torch.zeros(
                input_shape[:-1],
                dtype=torch.long,
                device=self.pos_ids.device,
            )
        tt_emb = self.token_type_embedding(token_type_ids)
        if self.use_token_type_embeddings:
            token_emb = token_emb + tt_emb

        pos_ids = self.pos_ids[:, : input_shape[1]]

        token_emb = token_emb + self.positional_embedding(pos_ids)
        token_emb = self.layer_norm(token_emb)
        token_emb = self.dropout(token_emb)
        return token_emb


class BaseTransformerEncoder(nn.Module):
    def __init__(
        self,
        num_layers: int,
        encoder_class: Callable,
        layerdrop_threshold: Optional[float] = 0.0,
        **kwargs,
    ) -> None:
        super(BaseTransformerEncoder, self).__init__()
        self.encoder_layers = nn.ModuleList([encoder_class(**kwargs) for _ in range(num_layers)])
        self.layerdrop_threshold = layerdrop_threshold

    def forward(
        self,
        hidden: torch.Tensor,
        attention_mask: Optional[torch.Tensor] = None,
        get_attention_scores: Optional[bool] = False,
    ) -> Union[Tuple[torch.Tensor], Tuple[torch.Tensor, torch.Tensor]]:
        attn_scores = []
        for layer in self.encoder_layers:
            # LayerDrop: https://arxiv.org/abs/1909.11556
            # TODO: make something with different shapes
            # of attn_scores while using LayerDrop
            dropout_probability = random.uniform(0, 1)
            if self.training and (dropout_probability < self.layerdrop_threshold):
                continue
            hidden = layer(
                hidden,
                attention_mask=attention_mask,
                get_attention_scores=get_attention_scores,
            )
            if get_attention_scores:
                attn_scores.append(hidden[1])
            hidden = hidden[0]
        output = (hidden,)
        if get_attention_scores:
            attn_scores = torch.stack(attn_scores, dim=-1)
            output = output + (attn_scores,)
        return output


class BaseTransformerDecoder(nn.Module):
    def __init__(
        self,
        num_layers: int,
        decoder_class: Callable,
        layerdrop_threshold: Optional[float] = 0.0,
        **kwargs,
    ) -> None:
        super(BaseTransformerDecoder, self).__init__()
        self.decoder_layers = nn.ModuleList([decoder_class(**kwargs) for _ in range(num_layers)])
        self.layerdrop_threshold = layerdrop_threshold

    def forward(
        self,
        hidden: torch.Tensor,
        encoder_hidden_state: Union[torch.Tensor, Tuple[torch.Tensor, torch.Tensor]],
        attention_mask: Optional[torch.Tensor] = None,
        encoder_attention_mask: Optional[Union[torch.Tensor, Tuple[torch.Tensor, torch.Tensor]]] = None,
        get_attention_scores: Optional[bool] = False,
    ) -> Union[Tuple[torch.Tensor], Tuple[torch.Tensor, torch.Tensor]]:
        attn_scores = []
        for layer in self.decoder_layers:
            # LayerDrop: https://arxiv.org/abs/1909.11556
            # TODO: make something with different shapes
            # of attn_scores while using LayerDrop
            dropout_probability = random.uniform(0, 1)
            if self.training and (dropout_probability < self.layerdrop_threshold):
                continue
            hidden = layer(
                hidden=hidden,
                encoder_hidden_state=encoder_hidden_state,
                attention_mask=attention_mask,
                encoder_attention_mask=encoder_attention_mask,
                get_attention_scores=get_attention_scores,
            )
            if get_attention_scores:
                attn_scores.append(hidden[1])
            hidden = hidden[0]

        output = (hidden,)
        if get_attention_scores:
            attn_scores = torch.stack(attn_scores, dim=-1)
            output = output + (attn_scores,)
        return output



# Copyright 2021 c00k1ez (https://github.com/c00k1ez). All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch

from .base_sampler import BaseSampler


class GreedySampler(BaseSampler):
    def __init__(self, *args, **kwargs) -> None:
        super(GreedySampler, self).__init__(*args, **kwargs)

    @torch.no_grad()
    def sample(self, logits: torch.Tensor, temperatype: float, **kwargs) -> torch.Tensor:
        probs = torch.softmax(logits / temperatype, dim=-1)
        max_val = probs.max(dim=-1)[1].unsqueeze(-1)
        return max_val



# Copyright 2021 c00k1ez (https://github.com/c00k1ez). All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

import torch
import torch.nn as nn
import torch.nn.functional as F


class LabelSmoothingLoss(nn.Module):
    def __init__(
        self, smoothing: Optional[float] = 0.0, ignore_index: Optional[int] = 0, reduction: Optional[str] = "mean"
    ) -> None:
        """
        ignore index must be equivalent to padding id
        """
        super(LabelSmoothingLoss, self).__init__()
        assert ignore_index >= 0
        assert reduction in ["mean", "sum", "none"]
        self.smooth = smoothing
        self.ignore_index = ignore_index
        self.reduction = reduction

    def forward(self, input: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        if len(input.shape) > 2:
            input = input.view(-1, input.shape[-1])
            target = target.view(-1)
        log_prob = F.log_softmax(input, dim=-1)
        weight = input.new_ones(input.shape) * self.smooth / (input.shape[-1] - 2)
        weight.scatter_(-1, target.unsqueeze(-1), (1.0 - self.smooth))
        weight.masked_fill_((target == self.ignore_index).unsqueeze(-1), 0)
        loss = (-weight * log_prob).sum(dim=-1)
        if self.reduction == "mean":
            loss = loss.sum() / (target != self.ignore_index).sum()
        elif self.reduction == "sum":
            loss = loss.sum()
        return loss



# Copyright 2021 c00k1ez (https://github.com/c00k1ez). All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

import torch
import torch.nn as nn
import torch.nn.functional as F


class UnlikelihoodLoss(nn.Module):
    def __init__(
        self,
        alpha: float,
        ignore_index: Optional[int] = 0,
        context_type: Optional[str] = "full_context",
        custom_likelihood_loss: Optional[nn.Module] = None,
        reduction: Optional[str] = "mean",
    ) -> None:
        super(UnlikelihoodLoss, self).__init__()
        assert context_type in ["full_context", "sentence"]
        self.context_type = context_type
        assert ignore_index >= 0
        self.ignore_index = ignore_index
        assert reduction in ["sum", "mean", "none"]
        self.reduction = reduction
        self.alpha = alpha
        if custom_likelihood_loss is not None:
            self.custom_likelihood_loss = custom_likelihood_loss
            assert custom_likelihood_loss.ignore_index == ignore_index
            assert custom_likelihood_loss.reduction == reduction
        else:
            self.custom_likelihood_loss = nn.CrossEntropyLoss(ignore_index=ignore_index, reduction=reduction)

    def forward(self, input: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        if self.context_type == "sentence" and len(input.shape) > 2:
            raise ValueError(
                "If you use context_type 'sentence', you should pass 3-d input tensor and 2-d target tensor"
            )
        raw_logits = F.log_softmax(input, dim=-1)
        unchanged_target = target.clone()
        with torch.no_grad():
            if self.context_type == "full_context":
                target = target.view(-1)
                raw_logits = raw_logits.view(-1, raw_logits.size(-1))
                negative_candidates = target.unsqueeze(0).expand(target.size(0), -1)
                mask = torch.empty_like(negative_candidates).fill_(self.ignore_index).triu()
                negative_candidates = negative_candidates.tril(-1) + mask
            elif self.context_type == "sentence":
                negative_candidates = target.unsqueeze(1).expand(-1, target.size(1), -1)
                mask = torch.empty_like(negative_candidates).fill_(self.ignore_index).triu()
                negative_candidates = negative_candidates.tril(-1) + mask

            negative_candidates.masked_fill_(negative_candidates == target.unsqueeze(-1), self.ignore_index)
            negative_targets = torch.zeros_like(raw_logits).scatter_(-1, negative_candidates, 1)
            # delete paddings from negative_targets
            negative_targets.masked_fill_(target.unsqueeze(-1) == self.ignore_index, 0)
            negative_targets[..., self.ignore_index] = 0.0

        # calculate unlikelihood part
        inverse_probs = torch.clamp((1.0 - raw_logits.exp()), min=1e-5)
        ul_loss = -torch.log(inverse_probs) * negative_targets
        if len(ul_loss.shape) > 2:
            ul_loss = ul_loss.view(-1, ul_loss.shape[-1])
        ul_loss = ul_loss.sum(dim=-1)

        # calculate mle loss
        mle_loss = self.custom_likelihood_loss(input.view(-1, input.shape[-1]), unchanged_target.view(-1))
        if self.reduction == "sum":
            ul_loss = ul_loss.sum()
        elif self.reduction == "mean":
            ul_loss = ul_loss.sum() / (target.view(-1) != self.ignore_index).sum()
        assert mle_loss.shape == ul_loss.shape
        return self.alpha * ul_loss + mle_loss



# Copyright 2021 c00k1ez (https://github.com/c00k1ez). All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial
from typing import Callable, Optional

import torch
import torch.nn as nn


def initialize_weights(
    model: nn.Module,
    initialization_function: Callable,
    pre_initialization: Optional[Callable] = None,
    is_admin=False,
    **kwargs,
) -> None:
    pre_init_outputs = {}
    if pre_initialization is not None:
        pre_init_outputs = pre_initialization(model)  # must be dict
    initialization_function_with_args = partial(initialization_function, **kwargs, **pre_init_outputs)
    model.apply(initialization_function_with_args)


def normal_initialization(module: nn.Module, init_range: Optional[float] = 0.02) -> None:
    if isinstance(module, (nn.Linear, nn.Embedding)):
        module.weight.data.normal_(mean=0.0, std=init_range)
        if isinstance(module, nn.Linear):
            if module.bias is not None:
                module.bias.data.zero_()
    elif isinstance(module, nn.LayerNorm):
        if module.bias is not None:
            module.bias.data.zero_()
        module.weight.data.fill_(1.0)



# Copyright 2021 c00k1ez (https://github.com/c00k1ez). All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, Tuple, Union

import torch
import torch.nn as nn
import torch.nn.functional as F

from .common_layers import BaseTransformerEncoder, TransformerEmbedding
from .post_ln_encoder import PostLNEncoderLayer
from .pre_ln_encoder import PreLNEncoderLayer
from .utils import create_attention_mask


class TransformerEncoder(nn.Module):
    def __init__(
        self,
        d_model: int,
        vocab_size: int,
        max_length: int,
        pad_token_id: int,
        token_type_vocab_size: int,
        n_heads: int,
        dim_feedforward: int,
        num_layers: int,
        dropout: Optional[float] = 0.1,
        use_embedding_layer_norm: Optional[bool] = False,
        pos_embedding_type: Optional[str] = "embedding",
        activation_name: Optional[str] = "gelu",
        ln_eps: Optional[float] = 1e-12,
        layerdrop_threshold: Optional[float] = 0.0,
        use_attention_merge_matrix: Optional[bool] = True,
        encoder_type: Optional[str] = "post_ln",
    ) -> None:
        super(TransformerEncoder, self).__init__()
        assert encoder_type in ["post_ln", "pre_ln"]
        self.encoder_type = encoder_type
        self.embedding = TransformerEmbedding(
            vocab_size=vocab_size,
            d_model=d_model,
            max_length=max_length,
            pad_token_id=pad_token_id,
            token_type_vocab_size=token_type_vocab_size,
            pos_embedding_type=pos_embedding_type,
            dropout=dropout,
            use_layer_norm=use_embedding_layer_norm,
            ln_eps=ln_eps,
        )
        encoder_class = PostLNEncoderLayer
        if encoder_type == "pre_ln":
            encoder_class = PreLNEncoderLayer
        self.encoder = BaseTransformerEncoder(
            num_layers=num_layers,
            encoder_class=encoder_class,
            d_model=d_model,
            n_heads=n_heads,
            dim_feedforward=dim_feedforward,
            dropout=dropout,
            activation_name=activation_name,
            ln_eps=ln_eps,
            layerdrop_threshold=layerdrop_threshold,
            use_attention_merge_matrix=use_attention_merge_matrix,
        )
        if encoder_type == "pre_ln":
            self.post_encoder_ln = nn.LayerNorm(d_model, eps=ln_eps)
        else:
            self.post_encoder_ln = nn.Identity()
            self.post_encoder_ln.register_parameter("weight", None)
            self.post_encoder_ln.register_parameter("bias", None)

    def forward(
        self,
        input_ids: torch.Tensor,
        attention_mask: Optional[torch.Tensor] = None,
        token_type_ids: Optional[torch.Tensor] = None,
        get_attention_scores: Optional[bool] = False,
    ) -> Union[Tuple[torch.Tensor], Tuple[torch.Tensor, torch.Tensor]]:

        attention_mask = create_attention_mask(attention_mask, input_ids.shape, input_ids.device)
        embeddings = self.embedding(input_ids, token_type_ids)
        hidden = self.encoder(
            embeddings,
            attention_mask=attention_mask,
            get_attention_scores=get_attention_scores,
        )
        hidden_ln = self.post_encoder_ln(hidden[0])
        output = (hidden_ln,)
        if get_attention_scores:
            output = output + (hidden[1],)
        return output



# Copyright 2021 c00k1ez (https://github.com/c00k1ez). All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, Tuple, Union

import torch
import torch.nn as nn
import torch.nn.functional as F

from .common_layers import FFN, BaseTransformerDecoder, MultiHeadAttention, TransformerEmbedding
from .utils import create_attention_mask


class PostLNDecoderLayer(nn.Module):
    def __init__(
        self,
        d_model: int,
        n_heads: int,
        dim_feedforward: int,
        dropout: Optional[float] = 0.1,
        activation_name: Optional[str] = "gelu",
        ln_eps: Optional[float] = 1e-12,
        context_len: Optional[int] = 512,
        use_attention_merge_matrix: Optional[bool] = True,
    ) -> None:
        super(PostLNDecoderLayer, self).__init__()
        self.self_attention = MultiHeadAttention(
            d_model=d_model,
            n_heads=n_heads,
            dropout=dropout,
            context_len=context_len,
        )
        if use_attention_merge_matrix:
            self.self_attn_merge_matrix = nn.Linear(d_model, d_model)
        else:
            self.self_attn_merge_matrix = nn.Identity()
            self.self_attn_merge_matrix.register_parameter("weight", None)
            self.self_attn_merge_matrix.register_parameter("bias", None)
        self.post_attn_ln = nn.LayerNorm(d_model, eps=ln_eps)

        self.cross_attention = MultiHeadAttention(d_model=d_model, n_heads=n_heads, dropout=dropout)
        if use_attention_merge_matrix:
            self.cross_attn_merge_matrix = nn.Linear(d_model, d_model)
        else:
            self.cross_attn_merge_matrix = nn.Identity()
            self.cross_attn_merge_matrix.register_parameter("weight", None)
            self.cross_attn_merge_matrix.register_parameter("bias", None)
        self.post_cross_attn_ln = nn.LayerNorm(d_model, eps=ln_eps)

        self.ffn = FFN(
            d_model=d_model,
            dim_feedforward=dim_feedforward,
            dropout=dropout,
            activation_name=activation_name,
        )
        self.post_ffn_ln = nn.LayerNorm(d_model, eps=ln_eps)

    def forward(
        self,
        hidden: torch.Tensor,
        encoder_hidden_state: torch.Tensor,
        attention_mask: Optional[torch.Tensor] = None,
        encoder_attention_mask: Optional[torch.Tensor] = None,
        get_attention_scores: Optional[bool] = False,
    ) -> Union[Tuple[torch.Tensor], Tuple[torch.Tensor, torch.Tensor]]:
        attn_scores = []
        self_attn_block = self.self_attention(
            query=hidden,
            key=hidden,
            value=hidden,
            attention_mask=attention_mask,
            get_attention_scores=get_attention_scores,
        )
        if get_attention_scores:
            attn_scores = [
                self_attn_block[1],
            ]
        self_attn_block = self_attn_block[0]
        self_attn_block = self.self_attn_merge_matrix(self_attn_block)
        self_attn_block = self_attn_block + hidden
        self_attn_block = self.post_attn_ln(self_attn_block)

        cross_attn_block = self.cross_attention(
            query=self_attn_block,
            key=encoder_hidden_state["key"],
            value=encoder_hidden_state["value"],
            attention_mask=encoder_attention_mask,
            get_attention_scores=get_attention_scores,
        )
        if get_attention_scores:
            attn_scores.append(cross_attn_block[1])
            attn_scores = torch.stack(attn_scores, dim=-1)
        cross_attn_block = cross_attn_block[0]
        cross_attn_block = self.cross_attn_merge_matrix(cross_attn_block)
        cross_attn_block = cross_attn_block + self_attn_block
        cross_attn_block = self.post_cross_attn_ln(cross_attn_block)

        ffn_block = self.ffn(cross_attn_block)
        ffn_block = ffn_block + cross_attn_block
        ffn_block = self.post_ffn_ln(ffn_block)

        output = (ffn_block,)
        if get_attention_scores:
            output = output + (attn_scores,)
        return output


class PostLNMultimodalDecoderLayer(nn.Module):
    def __init__(
        self,
        d_model: int,
        n_heads: int,
        dim_feedforward: int,
        dropout: Optional[float] = 0.1,
        activation_name: Optional[str] = "gelu",
        ln_eps: Optional[float] = 1e-12,
        context_len: Optional[int] = 512,
        use_attention_merge_matrix: Optional[bool] = True,
    ) -> None:
        super(PostLNMultimodalDecoderLayer, self).__init__()
        self.self_attention = MultiHeadAttention(
            d_model=d_model,
            n_heads=n_heads,
            dropout=dropout,
            context_len=context_len,
        )
        if use_attention_merge_matrix:
            self.self_attn_merge_matrix = nn.Linear(d_model, d_model)
        else:
            self.self_attn_merge_matrix = nn.Identity()
            self.self_attn_merge_matrix.register_parameter("weight", None)
            self.self_attn_merge_matrix.register_parameter("bias", None)

        self.post_attn_ln = nn.LayerNorm(d_model, eps=ln_eps)

        self.first_cross_attention = MultiHeadAttention(d_model=d_model, n_heads=n_heads, dropout=dropout)
        if use_attention_merge_matrix:
            self.first_cross_attn_merge_matrix = nn.Linear(d_model, d_model)
        else:
            self.first_cross_attn_merge_matrix = nn.Identity()
            self.first_cross_attn_merge_matrix.register_parameter("weight", None)
            self.first_cross_attn_merge_matrix.register_parameter("bias", None)
        self.first_post_cross_attn_ln = nn.LayerNorm(d_model, eps=ln_eps)

        self.second_cross_attention = MultiHeadAttention(d_model=d_model, n_heads=n_heads, dropout=dropout)
        if use_attention_merge_matrix:
            self.second_cross_attn_merge_matrix = nn.Linear(d_model, d_model)
        else:
            self.second_cross_attn_merge_matrix = nn.Identity()
            self.second_cross_attn_merge_matrix.register_parameter("weight", None)
            self.second_cross_attn_merge_matrix.register_parameter("bias", None)
        self.second_post_cross_attn_ln = nn.LayerNorm(d_model, eps=ln_eps)

        self.ffn = FFN(
            d_model=d_model,
            dim_feedforward=dim_feedforward,
            dropout=dropout,
            activation_name=activation_name,
        )
        self.post_ffn_ln = nn.LayerNorm(d_model, eps=ln_eps)

    def forward(
        self,
        hidden: torch.Tensor,
        encoder_hidden_state: Tuple[torch.Tensor, torch.Tensor],
        encoder_attention_mask: Tuple[torch.Tensor, torch.Tensor],
        attention_mask: Optional[torch.Tensor] = None,
        get_attention_scores: Optional[bool] = False,
    ) -> Union[Tuple[torch.Tensor], Tuple[torch.Tensor, torch.Tensor]]:
        (first_encoder_hidden_state, second_encoder_hidden_state) = (
            encoder_hidden_state[0],
            encoder_hidden_state[1],
        )
        (first_encoder_attention_mask, second_encoder_attention_mask) = (
            encoder_attention_mask[0],
            encoder_attention_mask[1],
        )
        attn_scores = []
        self_attn_block = self.self_attention(
            query=hidden,
            key=hidden,
            value=hidden,
            attention_mask=attention_mask,
            get_attention_scores=get_attention_scores,
        )
        if get_attention_scores:
            attn_scores = [
                self_attn_block[1],
            ]
        self_attn_block = self_attn_block[0]
        self_attn_block = self.self_attn_merge_matrix(self_attn_block)
        self_attn_block = self_attn_block + hidden
        self_attn_block = self.post_attn_ln(self_attn_block)

        first_cross_attn_block = self.first_cross_attention(
            query=self_attn_block,
            key=first_encoder_hidden_state["key"],
            value=first_encoder_hidden_state["value"],
            attention_mask=first_encoder_attention_mask,
            get_attention_scores=get_attention_scores,
        )
        if get_attention_scores:
            attn_scores.append(first_cross_attn_block[1])

        first_cross_attn_block = first_cross_attn_block[0]
        first_cross_attn_block = self.first_cross_attn_merge_matrix(first_cross_attn_block)
        first_cross_attn_block = first_cross_attn_block + self_attn_block
        first_cross_attn_block = self.first_post_cross_attn_ln(first_cross_attn_block)

        second_cross_attn_block = self.second_cross_attention(
            query=first_cross_attn_block,
            key=second_encoder_hidden_state["key"],
            value=second_encoder_hidden_state["value"],
            attention_mask=second_encoder_attention_mask,
            get_attention_scores=get_attention_scores,
        )
        if get_attention_scores:
            attn_scores.append(second_cross_attn_block[1])
            attn_scores = torch.stack(attn_scores, dim=-1)
        second_cross_attn_block = second_cross_attn_block[0]
        second_cross_attn_block = self.second_cross_attn_merge_matrix(second_cross_attn_block)
        second_cross_attn_block = second_cross_attn_block + first_cross_attn_block
        second_cross_attn_block = self.second_post_cross_attn_ln(second_cross_attn_block)

        ffn_block = self.ffn(second_cross_attn_block)
        ffn_block = ffn_block + second_cross_attn_block
        ffn_block = self.post_ffn_ln(ffn_block)

        output = (ffn_block,)
        if get_attention_scores:
            output = output + (attn_scores,)
        return output


class PostLNMultimodalTransformerDecoder(nn.Module):
    def __init__(
        self,
        d_model: int,
        vocab_size: int,
        max_length: int,
        pad_token_id: int,
        token_type_vocab_size: int,
        n_heads: int,
        dim_feedforward: int,
        num_layers: int,
        dropout: Optional[float] = 0.1,
        use_embedding_layer_norm: Optional[bool] = False,
        pos_embedding_type: Optional[str] = "embedding",
        activation_name: Optional[str] = "gelu",
        ln_eps: Optional[float] = 1e-12,
        layerdrop_threshold: Optional[float] = 0.0,
    ) -> None:
        super(PostLNMultimodalTransformerDecoder, self).__init__()
        self.embedding = TransformerEmbedding(
            vocab_size=vocab_size,
            d_model=d_model,
            max_length=max_length,
            pad_token_id=pad_token_id,
            token_type_vocab_size=token_type_vocab_size,
            dropout=dropout,
            use_layer_norm=use_embedding_layer_norm,
            ln_eps=ln_eps,
            pos_embedding_type=pos_embedding_type,
        )

        self.decoder = BaseTransformerDecoder(
            num_layers=num_layers,
            decoder_class=PostLNMultimodalDecoderLayer,
            d_model=d_model,
            n_heads=n_heads,
            dim_feedforward=dim_feedforward,
            dropout=dropout,
            activation_name=activation_name,
            ln_eps=ln_eps,
            context_len=max_length,
            layerdrop_threshold=layerdrop_threshold,
        )

    def forward(
        self,
        input_ids: torch.Tensor,
        first_encoder_hidden_state: torch.Tensor,
        second_encoder_hidden_state: torch.Tensor,
        attention_mask: Optional[torch.Tensor] = None,
        first_encoder_attention_mask: Optional[torch.Tensor] = None,
        second_encoder_attention_mask: Optional[torch.Tensor] = None,
        get_attention_scores: Optional[bool] = False,
    ) -> Union[Tuple[torch.Tensor], Tuple[torch.Tensor, torch.Tensor]]:
        attention_mask = create_attention_mask(
            attention_mask=attention_mask,
            input_shape=input_ids.shape,
            device=input_ids.device,
        )

        first_encoder_attention_mask = create_attention_mask(
            attention_mask=first_encoder_attention_mask,
            input_shape=first_encoder_hidden_state["key"].shape[:-1],
            device=first_encoder_hidden_state["key"].device,
            src_size=input_ids.shape[-1],
        )

        second_encoder_attention_mask = create_attention_mask(
            attention_mask=second_encoder_attention_mask,
            input_shape=second_encoder_hidden_state["key"].shape[:-1],
            device=second_encoder_hidden_state["key"].device,
            src_size=input_ids.shape[-1],
        )

        embeddings = self.embedding(input_ids)

        hidden = self.decoder(
            hidden=embeddings,
            encoder_hidden_state=(
                first_encoder_hidden_state,
                second_encoder_hidden_state,
            ),
            attention_mask=attention_mask,
            encoder_attention_mask=(
                first_encoder_attention_mask,
                second_encoder_attention_mask,
            ),
            get_attention_scores=get_attention_scores,
        )

        return hidden



# Copyright 2021 c00k1ez (https://github.com/c00k1ez). All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, Tuple, Union

import torch
import torch.nn as nn
import youtokentome as yttm
from plain_transformers import BPEWrapper


class BaseSampler(object):
    def __init__(
        self,
        model: nn.Module,
        encoder_tokenizer: Union[BPEWrapper, Tuple[BPEWrapper, BPEWrapper]],
        decoder_tokenizer: BPEWrapper,
        share_encoder_tokenizers: Optional[bool] = False,
        share_encoder_decoder_tokenizers: Optional[bool] = False,
    ) -> None:
        self.model = model
        self.first_encoder_tokenizer = None
        self.second_encoder_tokenizer = None
        self.decoder_tokenizer = decoder_tokenizer
        if isinstance(encoder_tokenizer, tuple):
            (
                self.first_encoder_tokenizer,
                self.second_encoder_tokenizer,
            ) = encoder_tokenizer
        else:
            self.first_encoder_tokenizer = encoder_tokenizer
        if share_encoder_tokenizers:
            self.first_encoder_tokenizer = self.second_encoder_tokenizer
        if share_encoder_decoder_tokenizers:
            self.decoder_tokenizer = self.first_encoder_tokenizer

    @torch.no_grad()
    def sample(self, logits: torch.Tensor, temperatype: float, **kwargs) -> torch.Tensor:
        raise NotImplementedError("You have to implement this method")

    @torch.no_grad()
    def generate(
        self,
        input_text,
        temperatype=1.0,
        max_length=50,
        second_input_text=None,
        decoder_input_text=None,
        device=None,
        **kwargs,
    ):
        if device is None:
            device = torch.device("cpu")
        if isinstance(input_text, torch.Tensor):
            assert len(input_text.shape) == 2
            encoder_input_ids = input_text.to(device)
        else:
            encoder_input_ids = torch.LongTensor(
                self.first_encoder_tokenizer.encode(
                    [input_text],
                    bos=True,
                    eos=True,
                    output_type=yttm.OutputType.ID,
                )
            ).to(device)
        second_encoder_input_ids = None
        if second_input_text is not None:
            if isinstance(second_input_text, torch.Tensor):
                assert len(second_input_text.shape) == 2
                second_encoder_input_ids = second_input_text.to(device)
            else:
                second_encoder_input_ids = torch.LongTensor(
                    self.second_encoder_tokenizer.encode(
                        [second_input_text],
                        bos=True,
                        eos=True,
                        output_type=yttm.OutputType.ID,
                    )
                ).to(device)

        labels = None
        if decoder_input_text is not None:
            if isinstance(decoder_input_text, torch.Tensor):
                assert len(decoder_input_text.shape) == 2
                labels = decoder_input_text.to(device)
            else:
                labels = torch.LongTensor(
                    self.decoder_tokenizer.encode(
                        [decoder_input_text],
                        bos=True,
                        eos=False,
                        output_type=yttm.OutputType.ID,
                    )
                ).to(device)
        else:
            labels = torch.LongTensor(
                [
                    [
                        self.decoder_tokenizer.bos_id,
                    ]
                ]
            ).to(device)

        args = {
            "get_logits": True,
            "return_encoder_state": True,
            "cached_encoder_state": None,
            "labels": labels,
        }
        if second_input_text is not None:
            args = {
                **args,
                "first_encoder_input_ids": encoder_input_ids,
                "second_encoder_input_ids": second_encoder_input_ids,
            }
        else:
            args = {
                **args,
                "input_ids": encoder_input_ids,
            }

        get_cached_encoder_state = False
        while args["labels"].shape[-1] < max_length:
            model_out = self.model(**args)
            logits = model_out["lm_probs"][:, -1, :]
            next_token = self.sample(logits, temperatype, **kwargs)
            if not get_cached_encoder_state:
                args["cached_encoder_state"] = model_out["encoder_hidden_state"]
                args["return_encoder_state"] = False
                get_cached_encoder_state = True
            args["labels"] = torch.cat([args["labels"], next_token], dim=-1)
        generated_seq = self.decoder_tokenizer.decode(
            args["labels"].cpu().tolist()[0],
            ignore_ids=(
                self.decoder_tokenizer.pad_id,
                self.decoder_tokenizer.unk_id,
                self.decoder_tokenizer.bos_id,
                self.decoder_tokenizer.eos_id,
            ),
        )[0]

        return generated_seq



# Copyright 2021 c00k1ez (https://github.com/c00k1ez). All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

import youtokentome as yttm


class BPEWrapper(yttm.BPE):
    def __init__(
        self,
        pad_id: Optional[int] = 0,
        unk_id: Optional[int] = 1,
        bos_id: Optional[int] = 2,
        eos_id: Optional[int] = 3,
        *args,
        **kwargs,
    ) -> None:
        super(BPEWrapper, self).__init__(*args, **kwargs)
        assert self.vocab()[pad_id] == "<PAD>"
        assert self.vocab()[unk_id] == "<UNK>"
        assert self.vocab()[bos_id] == "<BOS>"
        assert self.vocab()[eos_id] == "<EOS>"
        self.pad_id = pad_id
        self.unk_id = unk_id
        self.bos_id = bos_id
        self.eos_id = eos_id



# Copyright 2021 c00k1ez (https://github.com/c00k1ez). All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, Tuple, Union

import torch
import torch.nn as nn
import torch.nn.functional as F

from .common_layers import BaseTransformerDecoder, TransformerEmbedding
from .post_ln_decoder import PostLNDecoderLayer, PostLNMultimodalDecoderLayer
from .utils import create_attention_mask


# TODO: implement pre ln decoder
class TransformerDecoder(nn.Module):
    def __init__(
        self,
        d_model: int,
        vocab_size: int,
        max_length: int,
        pad_token_id: int,
        token_type_vocab_size: int,
        n_heads: int,
        dim_feedforward: int,
        num_layers: int,
        dropout: Optional[float] = 0.1,
        use_embedding_layer_norm: Optional[bool] = False,
        pos_embedding_type: Optional[str] = "embedding",
        activation_name: Optional[str] = "gelu",
        ln_eps: Optional[float] = 1e-12,
        layerdrop_threshold: Optional[float] = 0.0,
        use_attention_merge_matrix: Optional[bool] = True,
        decoder_type: Optional[str] = "post_ln",
    ) -> None:
        super(TransformerDecoder, self).__init__()
        self.embedding = TransformerEmbedding(
            vocab_size=vocab_size,
            d_model=d_model,
            max_length=max_length,
            pad_token_id=pad_token_id,
            token_type_vocab_size=token_type_vocab_size,
            dropout=dropout,
            use_layer_norm=use_embedding_layer_norm,
            ln_eps=ln_eps,
            pos_embedding_type=pos_embedding_type,
        )

        self.decoder = BaseTransformerDecoder(
            num_layers=num_layers,
            decoder_class=PostLNDecoderLayer,
            d_model=d_model,
            n_heads=n_heads,
            dim_feedforward=dim_feedforward,
            dropout=dropout,
            activation_name=activation_name,
            ln_eps=ln_eps,
            context_len=max_length,
            layerdrop_threshold=layerdrop_threshold,
            use_attention_merge_matrix=use_attention_merge_matrix,
        )

    def forward(
        self,
        input_ids: torch.Tensor,
        encoder_hidden_state: torch.Tensor,
        attention_mask: Optional[torch.Tensor] = None,
        encoder_attention_mask: Optional[torch.Tensor] = None,
        get_attention_scores: Optional[bool] = False,
    ) -> Union[Tuple[torch.Tensor], Tuple[torch.Tensor, torch.Tensor]]:

        attention_mask = create_attention_mask(
            attention_mask=attention_mask,
            input_shape=input_ids.shape,
            device=input_ids.device,
        )

        encoder_attention_mask = create_attention_mask(
            attention_mask=encoder_attention_mask,
            input_shape=encoder_hidden_state["key"].shape[:-1],
            device=encoder_hidden_state["key"].device,
            src_size=input_ids.shape[-1],
        )

        embeddings = self.embedding(input_ids)

        hidden = self.decoder(
            hidden=embeddings,
            encoder_hidden_state=encoder_hidden_state,
            attention_mask=attention_mask,
            encoder_attention_mask=encoder_attention_mask,
            get_attention_scores=get_attention_scores,
        )

        return hidden


class MultimodalTransformerDecoder(nn.Module):
    def __init__(
        self,
        d_model: int,
        vocab_size: int,
        max_length: int,
        pad_token_id: int,
        token_type_vocab_size: int,
        n_heads: int,
        dim_feedforward: int,
        num_layers: int,
        dropout: Optional[float] = 0.1,
        use_embedding_layer_norm: Optional[bool] = False,
        pos_embedding_type: Optional[str] = "embedding",
        activation_name: Optional[str] = "gelu",
        ln_eps: Optional[float] = 1e-12,
        layerdrop_threshold: Optional[float] = 0.0,
        use_attention_merge_matrix: Optional[bool] = True,
        decoder_type: Optional[str] = "post_ln",
    ) -> None:
        super(MultimodalTransformerDecoder, self).__init__()
        self.embedding = TransformerEmbedding(
            vocab_size=vocab_size,
            d_model=d_model,
            max_length=max_length,
            pad_token_id=pad_token_id,
            token_type_vocab_size=token_type_vocab_size,
            dropout=dropout,
            use_layer_norm=use_embedding_layer_norm,
            ln_eps=ln_eps,
            pos_embedding_type=pos_embedding_type,
        )

        self.decoder = BaseTransformerDecoder(
            num_layers=num_layers,
            decoder_class=PostLNMultimodalDecoderLayer,
            d_model=d_model,
            n_heads=n_heads,
            dim_feedforward=dim_feedforward,
            dropout=dropout,
            activation_name=activation_name,
            ln_eps=ln_eps,
            context_len=max_length,
            layerdrop_threshold=layerdrop_threshold,
            use_attention_merge_matrix=use_attention_merge_matrix,
        )

    def forward(
        self,
        input_ids: torch.Tensor,
        first_encoder_hidden_state: torch.Tensor,
        second_encoder_hidden_state: torch.Tensor,
        attention_mask: Optional[torch.Tensor] = None,
        first_encoder_attention_mask: Optional[torch.Tensor] = None,
        second_encoder_attention_mask: Optional[torch.Tensor] = None,
        get_attention_scores: Optional[bool] = False,
    ) -> Union[Tuple[torch.Tensor], Tuple[torch.Tensor, torch.Tensor]]:
        attention_mask = create_attention_mask(
            attention_mask=attention_mask,
            input_shape=input_ids.shape,
            device=input_ids.device,
        )

        first_encoder_attention_mask = create_attention_mask(
            attention_mask=first_encoder_attention_mask,
            input_shape=first_encoder_hidden_state["key"].shape[:-1],
            device=first_encoder_hidden_state["key"].device,
            src_size=input_ids.shape[-1],
        )

        second_encoder_attention_mask = create_attention_mask(
            attention_mask=second_encoder_attention_mask,
            input_shape=second_encoder_hidden_state["key"].shape[:-1],
            device=second_encoder_hidden_state["key"].device,
            src_size=input_ids.shape[-1],
        )

        embeddings = self.embedding(input_ids)

        hidden = self.decoder(
            hidden=embeddings,
            encoder_hidden_state=(
                first_encoder_hidden_state,
                second_encoder_hidden_state,
            ),
            attention_mask=attention_mask,
            encoder_attention_mask=(
                first_encoder_attention_mask,
                second_encoder_attention_mask,
            ),
            get_attention_scores=get_attention_scores,
        )

        return hidden



# Copyright 2021 c00k1ez (https://github.com/c00k1ez). All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, Tuple, Union

import torch
import torch.nn as nn
import torch.nn.functional as F

from .common_layers import FFN, MultiHeadAttention


class PreLNEncoderLayer(nn.Module):
    def __init__(
        self,
        d_model: int,
        n_heads: int,
        dim_feedforward: int,
        dropout: Optional[float] = 0.1,
        activation_name: Optional[str] = "gelu",
        ln_eps: Optional[float] = 1e-12,
        use_attention_merge_matrix: Optional[bool] = True,
    ) -> None:
        super(PreLNEncoderLayer, self).__init__()
        self.self_attention = MultiHeadAttention(d_model, n_heads, dropout)
        self.dropout = nn.Dropout(dropout)
        self.ffn = FFN(d_model, dim_feedforward, dropout, activation_name=activation_name)
        self.pre_attn_ln = nn.LayerNorm(d_model, eps=ln_eps)
        self.pre_ffn_ln = nn.LayerNorm(d_model, eps=ln_eps)
        if use_attention_merge_matrix:
            self.merge_matrix = nn.Linear(d_model, d_model)
        else:
            self.merge_matrix = nn.Identity()
            self.merge_matrix.register_parameter("weight", None)
            self.merge_matrix.register_parameter("bias", None)

    def forward(
        self,
        hidden: torch.Tensor,
        attention_mask: Optional[torch.Tensor] = None,
        get_attention_scores: Optional[bool] = False,
    ) -> Union[Tuple[torch.Tensor], Tuple[torch.Tensor, torch.Tensor]]:

        attn_scores = None
        block_state = self.pre_attn_ln(hidden)
        block_state = self.self_attention(
            query=block_state,
            key=block_state,
            value=block_state,
            attention_mask=attention_mask,
            get_attention_scores=get_attention_scores,
        )
        if get_attention_scores:
            attn_scores = block_state[1]
        block_state = block_state[0]
        block_state = self.merge_matrix(block_state)
        block_state = block_state + hidden

        ffn_block_state = self.pre_ffn_ln(block_state)
        ffn_block_state = self.ffn(ffn_block_state)
        ffn_block_state = ffn_block_state + block_state

        output = (ffn_block_state,)
        if get_attention_scores:
            output = output + (attn_scores,)
        return output



# Copyright 2021 c00k1ez (https://github.com/c00k1ez). All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable, Tuple, Union

import torch
import torch.nn.functional as F


def create_attention_mask(
    attention_mask: Union[torch.Tensor, None],
    input_shape: Tuple[int, int],
    device: torch.device,
    src_size: int = 1,
) -> torch.Tensor:
    # [batch_size, seq_len] -> [batch_size, 1, src_size, seq_len]
    if attention_mask is None:
        attention_mask = torch.ones(*input_shape).to(device)
    attention_mask = attention_mask.unsqueeze(1).unsqueeze(1)
    if src_size > 1:
        attention_mask = attention_mask.repeat(1, 1, src_size, 1)
    attention_mask = (1.0 - attention_mask) * -10000.0
    return attention_mask


def act_to_func(act_name: str) -> Callable:
    acts = {
        "gelu": F.gelu,
        "relu": F.relu,
        "relu6": F.relu6,
        "elu": F.elu,
        "selu": F.selu,
        "celu": F.celu,
        "leaky_relu": F.leaky_relu,
        "tanh": F.tanh,
    }
    if act_name in acts:
        return acts[act_name]
    else:
        return F.relu



# Copyright 2021 c00k1ez (https://github.com/c00k1ez). All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import re

import setuptools


def version():
    current_dir = os.path.abspath(os.path.dirname(__file__))
    version_file = os.path.join(current_dir, "src/plain_transformers", "__init__.py")
    with open(version_file, encoding="utf-8") as f:
        return re.search(r'^__version__ = [\'"]([^\'"]*)[\'"]', f.read(), re.M).group(1)


with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()


def _load_requirements(path_dir: str, file_name: str = "requirements.txt", comment_char: str = "#"):
    with open(os.path.join(path_dir, file_name), "r") as file:
        lines = [ln.strip() for ln in file.readlines()]
    reqs = []
    for ln in lines:
        if comment_char in ln:
            ln = ln[: ln.index(comment_char)].strip()
        if ln.startswith("http"):
            continue
        if ln:
            reqs.append(ln)
    return reqs


setuptools.setup(
    name="plain-transformers",
    version=version(),
    author="c00k1ez",
    author_email="c00k1ez.th13f@gmail.com",
    description="one more transformers lib",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/c00k1ez/plain_transformers",
    project_urls={
        "Bug Tracker": "https://github.com/c00k1ez/plain_transformers/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src", exclude=["tests/", "tests/*"]),
    python_requires=">=3.6",
    install_requires=_load_requirements(os.path.dirname(__file__)),
    include_package_data=True,
)



# Copyright 2021 c00k1ez (https://github.com/c00k1ez). All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Optional

import torch
from torch.optim.lr_scheduler import _LRScheduler


class NoamScheduler(_LRScheduler):
    def __init__(
        self,
        optimizer: torch.optim.Optimizer,
        d_model: int,
        warmup_steps: Optional[int] = 4000,
    ) -> None:
        self.d_model = d_model
        self.warmup_steps = warmup_steps
        super(NoamScheduler, self).__init__(optimizer)

    def get_lr(self) -> List[float]:
        last_step = max(1, self.last_epoch)
        scale_factor = self.d_model ** (-0.5) * min(last_step ** (-0.5), last_step * self.warmup_steps ** (-1.5))
        new_lrs = [lr * scale_factor for lr in self.base_lrs]
        return new_lrs
